/*
	Physical Simulation in Graphics Scene - master's thesis
    Copyright (C) 2010-2011  Marin Javorka, FIT VUT in Brno, http://www.fit.vutbr.cz, marian.javorka@deq.sk

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/
#include <sstream>

#include <osgDB/ReadFile>
#include <osg/PositionAttitudeTransform>
#include <osgUtil/Optimizer>
#include <osgViewer/CompositeViewer>
#include <osgViewer/ViewerEventHandlers>

#include <osgGA/KeySwitchMatrixManipulator>

#include <btBulletDynamicsCommon.h>
#include "BulletDynamics/Vehicle/btRaycastVehicle.h"
#include "BulletCollision/CollisionShapes/btShapeHull.h"
#include "BulletCollision/CollisionShapes/btBvhTriangleMeshShape.h"


#include "physics.h"
#include "my_key_event_handler.h"
#include "hud.h"
#include "vehicle.h"
#include "camera.h"
#include "scene.h"
#include "menu.h"

#include "vector"
#include <osg/StateSet>

#include <osgParticle/PrecipitationEffect>

#include <winsock.h>
#include "mysql_driver.h"
#include "mysql_connection.h"
#include "cppconn/statement.h"
#include "cppconn/resultset.h"


// osg example class
class MyGustCallback : public osg::NodeCallback
{

    public:

        MyGustCallback() {}

        virtual void operator()(osg::Node* node, osg::NodeVisitor* nv)
        {
            osgParticle::PrecipitationEffect* pe = dynamic_cast<osgParticle::PrecipitationEffect*>(node);
            
            float value = sin(nv->getFrameStamp()->getSimulationTime());
            if (value<-0.5)
            {
                pe->snow(1.0);
            }
            else
            {
                pe->rain(0.5);
            }
        
            traverse(node, nv);
        }
};

// osg example class
class MotionBlurOperation: public osg::Operation
{
public:
   MotionBlurOperation(double persistence):
       osg::Operation("MotionBlur",true),
       cleared_(false),
       persistence_(persistence)
   {
   }
	
   virtual void operator () (osg::Object* object)
   {
       osg::GraphicsContext* gc = dynamic_cast<osg::GraphicsContext*>(object);
       if (!gc) return;
	    
       double t = gc->getState()->getFrameStamp()->getSimulationTime();
	
       if (!cleared_)
       {
           // clear the accumulation buffer
           glClearColor(0, 0, 0, 0);
           glClear(GL_ACCUM_BUFFER_BIT);
           cleared_ = true;
           t0_ = t;
       }
	
       double dt = fabs(t - t0_);
       t0_ = t;
	
       // compute the blur factor
       double s = powf(0.2, dt / persistence_);

	   // scale, accumulate and return
       glAccum(GL_MULT, s);
       glAccum(GL_ACCUM, 1 - s);
       glAccum(GL_RETURN, 1.0f);
   }

private:
    bool cleared_;
    double t0_;
    double persistence_;
};

int main(int argc, char* argv[])
{

	bool mysql_connected = false;
	int slow_connection = 0;

	// default values
	bool connect_to_mysql = false;
	int screen_num = 0;
	string cars_data[2];
	cars_data[0] = "data/Renault_megane_3/megane3";
	cars_data[1] = "data/Ferrari_f430/f430";
	int cars_used[2];
	string players_names[2];
	players_names[0] = "";
	players_names[1] = "";
	float players_colors[6];
	players_colors[0] = 0.9;
	players_colors[1] = 0.6;
	players_colors[2] = 0.0;
	players_colors[3] = 0.8;
	players_colors[4] = 0.0;
	players_colors[5] = 0.0;
	int weather = 3;
	const double max_intensity = 0.6;
	double intensity;
	int intense = 0;
	bool players_online[2];
	players_online[0] = false;
	players_online[1] = false;

	int automat[2];
	automat[0] = 1;
	automat[1] = 1;

	cars_used[0] = 3;
	cars_used[1] = 3;

	// inicialize some timers
    osg::Timer_t frame_tick = osg::Timer::instance()->tick();	// every frame tick

	srand(frame_tick);

	FILE * pFile;

	// look for settings file
	pFile = fopen ( "data/Brno/tmp" , "r" );
	if(pFile != NULL ){
		rewind (pFile);
		//0 name1 1 0 0.9 0.8 0.0 name2 1 0 0.8 0.0 0.0
		char name1[64] = "", name2[64] = "";
		int online1 = 0, online2 = 0;
		fscanf(pFile, "%d %d %s %d %d %d %f %f %f %s %d %d %d %f %f %f", 
			&weather, &slow_connection, name1, &online1, &cars_used[0], &automat[0], &players_colors[0], &players_colors[1], &players_colors[2],
			name2, &online2, &cars_used[1], &automat[1], &players_colors[3], &players_colors[4], &players_colors[5]);

		players_names[0] = name1;
		players_names[1] = name2;
		if(online1 == 1){
			players_online[0] = true;
			connect_to_mysql = true;
		}
		else
			players_online[0] = false;
		if(online2 == 1){
			players_online[1] = true;
			connect_to_mysql = true;
		}
		else
			players_online[1] = false;

		// terminate
		fclose (pFile);
	}

	if( cars_used[0] == 3 )
		cars_used[0] = (rand()%10)%2;
	if( cars_used[1] == 3 )
		cars_used[1] = (rand()%10)%2;

	// weather is random
	if(weather == 3){
		weather = rand()%2+3;
	}
	if(weather > 0){
		intensity = (rand()%(int)(max_intensity*100)+1)/(max_intensity*100.0);
		if( weather < 3 ){
			if(intensity <= 0.2)
				intense = 0;
			else if(intensity <= 0.4)
				intense = 1;
			else
				intense = 2;
		}
	}

	// init mysql connection
	sql::mysql::MySQL_Driver *driver;
	sql::Connection	*con;
	sql::Statement *stmt;

	if( connect_to_mysql == true ){
		mysql_connected = true;
		driver = sql::mysql::MySQL_Driver::Instance();
		con = driver->connect("echo.datacentre.sk", "c20348dip", "password");
		if ( con == NULL ) {
			cerr << con->getWarnings() << endl;
			mysql_connected = false;
		}
		else{
			stmt = con->createStatement();
			stmt->execute("USE c20348dip");
		}
	}
	else
		mysql_connected = false;


	// create WSI
    osg::GraphicsContext::WindowingSystemInterface* wsi = osg::GraphicsContext::getWindowingSystemInterface();
	// creation failed
    if (!wsi){
        osg::notify(osg::NOTICE)<<"Error, no WindowSystemInterface available, cannot create windows."<<std::endl;
        return 1;
    }

    unsigned int width, height;


	// get screen resolution
    wsi->getScreenResolution(osg::GraphicsContext::ScreenIdentifier(screen_num), width, height);

	// physics dynamic world
	btDynamicsWorld* physicalWorld = physics->initPhysics();

	// generate temperatures
	if(weather == 0)
		physics->setTemperatureC(rand()%21+20);
	else if(weather == 1)
		physics->setTemperatureC(rand()%21+5);
	else if(weather == 2)
		physics->setTemperatureC(rand()%16-10);
	else if(weather == 3)
		physics->setTemperatureC(rand()%21+12);
	else if(weather == 4)
		physics->setTemperatureC(rand()%16);

    // scenegraph scene
	osg::StateSet* skyStateSet;
	osg::ref_ptr<osg::Group> scene = loadTrackAndEnviroment(physicalWorld, weather, &skyStateSet);

	// generate sky
	if( weather == 0 )
		skyStateSet->addUniform(new osg::Uniform("alpha", 0.0f));
	else{
		float w_intensity = 0.5 + intensity;
		if( w_intensity > 1.0 )
			w_intensity = 1.0;

		skyStateSet->addUniform(new osg::Uniform("alpha", (float)w_intensity));
	}

	/////////////////////////
	// Vehicles
	/////////////////////////
	vector<Player*> players;

	// store info about players
	players.push_back(new Player());
	players.at(0)->path = cars_data[cars_used[0]];
	players.at(0)->lap_time = 0;
	players.at(0)->lapT = 0;
	players.at(0)->lap = 0;
	players.at(0)->bestLapT = -1;
	players.at(0)->bestLapTmp = -1;
	players.at(0)->bestS1 = -1;
	players.at(0)->bestS2 = -1;
	players.at(0)->recordLapT = -1;
	players.at(0)->recordS1 = -1;
	players.at(0)->recordS2 = -1;
	players.at(0)->sector1 = 0;
	players.at(0)->sector2 = 0;
	players.at(0)->sector3 = 0;
	players.at(0)->d_old = 10000;
	players.at(0)->d2_old = 10000;
	players.at(0)->d3_old = 10000;
	players.at(0)->oldPos = osg::Vec3d(0,-5,0.1);
	players.at(0)->name = players_names[0];
	players.at(0)->online = players_online[0];
	players.at(0)->color = osg::Vec3(players_colors[0],players_colors[1],players_colors[2]);
	if( automat[0] == 1 )
		players.at(0)->manual = false;
	else
		players.at(0)->manual = true;

	if( !players_names[1].empty() ){
		players.push_back(new Player());
		players.at(1)->path = cars_data[cars_used[1]];
		players.at(1)->lap_time = 0;
		players.at(1)->lapT = 0;
		players.at(1)->lap = 0;
		players.at(1)->bestLapT = -1;
		players.at(1)->bestLapTmp = -1;
		players.at(1)->bestS1 = -1;
		players.at(1)->bestS2 = -1;
		players.at(1)->recordLapT = -1;
		players.at(1)->recordS1 = -1;
		players.at(1)->recordS2 = -1;
		players.at(1)->sector1 = 0;
		players.at(1)->sector2 = 0;
		players.at(1)->sector3 = 0;
		players.at(1)->d_old = 10000;
		players.at(1)->d2_old = 10000;
		players.at(1)->d3_old = 10000;
		players.at(1)->oldPos = osg::Vec3d(0,-5,0.1);
		players.at(1)->name = players_names[1];
		players.at(1)->online = players_online[1];
		players.at(1)->color = osg::Vec3(players_colors[3],players_colors[4],players_colors[5]);
		if( automat[1] == 1 )
			players.at(1)->manual = false;
		else
			players.at(1)->manual = true;
	}

	// add players to scene
	addPlayers(players, physicalWorld, scene, width, height);


	if(mysql_connected == true){

		// load personal bests
		for(int i = 0; i < (int)players.size(); i++){
			if( players.at(i)->online == false )
				continue;
			sql::ResultSet *res;
			char query[1024];
			sprintf(query,"SELECT `lap_time`,`sector1`,`sector2` FROM `c20348dip`.`times` WHERE `name` = '%s' AND `car_type` = '%d' AND `weather` = '%d-%d' LIMIT 1",players.at(i)->name.c_str(), cars_used[i], weather, intense);
			res = stmt->executeQuery(query);

			while (res->next()){
				char lap_text[64];
				int mins;
				float secs;

				players.at(i)->bestLapT = res->getDouble("lap_time");
				players.at(i)->bestLapTmp = res->getDouble("lap_time");

				mins = players.at(i)->bestLapT/60;
				secs = players.at(i)->bestLapT - 60*mins;
				sprintf(lap_text,"%d:%06.3f",mins,secs);
				players.at(i)->hud->HUDBestLap->setText(lap_text);

				players.at(i)->bestS1 = res->getDouble("sector1");

				mins = players.at(i)->bestS1/60;
				secs = players.at(i)->bestS1 - 60*mins;
				sprintf(lap_text,"s1: %d:%06.3f",mins,secs);
				players.at(i)->hud->HUDBestLapS1->setText(lap_text);

				players.at(i)->bestS2 = res->getDouble("sector2");
			}

			// load track record
			sprintf(query,"SELECT `lap_time`,`name`,`sector1`,`sector2` FROM `c20348dip`.`times` WHERE `car_type` = '%d' AND `weather` = '%d-%d'  ORDER BY `lap_time` ASC LIMIT 1", cars_used[i], weather, intense);
			res = stmt->executeQuery(query); 

			while (res->next()){
				char lap_text[64],lapS1_text[64];
				int mins;
				float secs;

				mins = res->getDouble("lap_time")/60;
				secs = res->getDouble("lap_time") - 60*mins;
				sprintf(lap_text,"%d:%06.3f",mins,secs);
					
				mins = res->getDouble("sector1")/60;
				secs = res->getDouble("sector1") - 60*mins;
				sprintf(lapS1_text,"s1: %d:%06.3f",mins,secs);

				players.at(i)->recordLapT = res->getDouble("lap_time");
				players.at(i)->recordS1 = res->getDouble("sector1");
				players.at(i)->recordS2 = res->getDouble("sector2");
				players.at(i)->hud->HUDRecord->setText(lap_text);
				players.at(i)->hud->HUDRecordman->setText(res->getString("name"));
				players.at(i)->hud->HUDRecordS1->setText(lapS1_text);
			}

			delete res;
		}
	}

	/////////////////////////
	// Vehicles END
	/////////////////////////

	// create Key Event Handler
	osg::ref_ptr<MyKeyEventHandler> handler = keyEH;

	/////////////////////////
	// Viewer
	///////////////////////// 
	
	// set the display settings we can to request, OsgCameraGroup will read this.
	osg::DisplaySettings::instance()->setMinimumNumAccumBits(8,8,8,8);

	osg::ref_ptr<osgViewer::CompositeViewer> viewer = new osgViewer::CompositeViewer();

    osg::ref_ptr<osg::GraphicsContext::Traits> traits = new osg::GraphicsContext::Traits;
    traits->x = 0;
    traits->y = 0;
    traits->width = width;
    traits->height = height;
    traits->windowDecoration = false;
    traits->doubleBuffer = true;
    traits->sharedContext = 0;
	traits->screenNum = screen_num;
	traits->useCursor = false;
	traits->useMultiThreadedOpenGLEngine = true;
	traits->samples = 8;

	// create graphics context
    osg::ref_ptr<osg::GraphicsContext> gc = osg::GraphicsContext::createGraphicsContext(traits.get());
    if (gc.valid()){
		gc->setClearColor(osg::Vec4f(0.2f,0.2f,0.6f,1.0f));
		gc->setClearMask(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

	for(int i = 0; i < (int)players.size(); i++){
		players.at(i)->ksw = new osgGA::KeySwitchMatrixManipulator;

		// view
		{	
			// camera manipulator
			osg::Matrixd m;
			m.setTrans(0,-2,-20.0);
			//m.setTrans(0,0,-25);
			m.setRotate(osg::Quat(3.14/2.1,osg::Vec3f(-1,0,0)));
			players.at(i)->fc = new FollowCam(players.at(i)->carMT,m);
			
			players.at(i)->ksw->addMatrixManipulator('0', "Follow", players.at(i)->fc);

			// create view
			osgViewer::View* view = new osgViewer::View;
			view->setName("Player");
			viewer->addView(view);

			// set up camera
			view->getCamera()->setClearColor(osg::Vec4f(0.2f,0.2f,0.6f,1.0f));

			// change perspective
			double fovy, aspectRatio, zNear, zFar;
			view->getCamera()->getProjectionMatrixAsPerspective(fovy, aspectRatio, zNear, zFar);

			double newAspectRatio = double(width/players.size()) / double(height);
			double aspectRatioChange = newAspectRatio / aspectRatio;
			if (aspectRatioChange != 1.0)
			{
				view->getCamera()->getProjectionMatrix() *= osg::Matrix::scale(1.0/aspectRatioChange,1.0,1.0);
			}

			//view->getCamera()->setViewport(new osg::Viewport(traits->x, traits->y, traits->width, traits->height));
			int off_x = 1;
			if(players.size() == 1)
				off_x = 0;
			view->getCamera()->setViewport(new osg::Viewport((off_x-i)*width/players.size(),0, width/players.size(),height));
			GLenum buffer = traits->doubleBuffer ? GL_BACK : GL_FRONT;

			view->getCamera()->setDrawBuffer(buffer);
			view->getCamera()->setReadBuffer(buffer);

			view->getCamera()->setGraphicsContext(gc.get());
			view->setCameraManipulator(players.at(i)->ksw.get());

			view->setSceneData(players.at(i)->node);

			// add some handlers
			view->addEventHandler( new osgViewer::ThreadingHandler );
			view->addEventHandler(handler.get());
		}
	}

	/////////////////////////
	// Viewer END
	/////////////////////////



    // optimize the scene graph, remove redundant nodes and state etc.
	//osgUtil::Optimizer* optimizer = new osgUtil::Optimizer();
	//optimizer->optimize(scene.get());

	// set Anisotropic filtering 16x 
	osgUtil::Optimizer::TextureVisitor visitor(true,false,false,false,true,16.0);
	visitor.traverse(*(osg::Node*)scene.get());

	viewer->realize();

	// hiding cursor and setting up motion blur
	double persistence = 0.06;   
	osgViewer::ViewerBase::Windows windows; 
	viewer->getWindows(windows); 
	for (osgViewer::ViewerBase::Windows::iterator itr = windows.begin(); itr != windows.end(); ++itr){ 
		(*itr)->setCursor(osgViewer::GraphicsWindow::NoCursor); 
		(*itr)->add(new MotionBlurOperation(persistence));
	}


	bool pause = false;

	// set defaul tyre friction (real is about 0.8, but that causes massive understeer)
	double frict = 1.8;
	const double max_frict = 1.8;
	const double rain_coef = 0.4;
	const double snow_coef = 0.6;

	for(int i = 0; i < (int)players.size(); i++){
		players.at(i)->car->setFriction(0,frict);
		players.at(i)->car->setFriction(1,frict);
		players.at(i)->car->setFriction(2,frict);
		players.at(i)->car->setFriction(3,frict);
	}

	// rain and snow effect (inspired from osg demo)
	osg::ref_ptr<osgParticle::PrecipitationEffect> precipitationEffect = new osgParticle::PrecipitationEffect;
	if(weather > 0){

		scene->addChild(precipitationEffect.get());

		scene->getOrCreateStateSet()->setAttributeAndModes(precipitationEffect->getFog());

		if(weather == 1 || weather == 3)
			precipitationEffect->rain(intensity);
		else
			precipitationEffect->snow(intensity);

		precipitationEffect->setWind(osg::Vec3((rand()%8/10.0),(rand()%8/10.0),0));

		precipitationEffect->setUseFarLineSegments(true);

		if(weather == 1 || weather == 3)
			frict -= (intensity/max_intensity)*rain_coef;
		else
			frict -= (intensity/max_intensity)*snow_coef;

		for(int i = 0; i < (int)players.size(); i++){
			players.at(i)->car->setFriction(0,frict);
			players.at(i)->car->setFriction(1,frict);
			players.at(i)->car->setFriction(2,frict);
			players.at(i)->car->setFriction(3,frict);
		}

	}


	double temperature[4] = {0.8f,0.8f,0.8f,0.8f};

	osg::Timer_t start_brake = 0;
	osg::Timer_t start_hbrake = 0;

	osg::Timer_t weather_timer = osg::Timer::instance()->tick();
	bool weather_changed = false;
	bool intensity_changed = false;
	float intensity_inc = 0.0f;

	int weather_change_time = 60;
	
	// loop for every frame
    while (!viewer->done()){	

		// actual tick
		osg::Timer_t now_tick = osg::Timer::instance()->tick();

		float dt = osg::Timer::instance()->delta_s(frame_tick, now_tick);

		// dynamic weather changes
		if( weather > 2 ){
			if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % weather_change_time == 0 && weather_changed == false ){
				weather_changed = true;
				srand(now_tick);
				float final_intensity = (rand()%(int)(max_intensity*100)+1)/(max_intensity*100.0);
				intensity_inc = (final_intensity - intensity)/float(weather_change_time/2.0);
			}
			else if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % weather_change_time > 0 && weather_changed == true ){
				weather_changed = false;
				srand(now_tick);
				weather_change_time = rand()%60+60;
				//weather_change_time = 10;
			}

			if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % 2 == 0 && intensity_changed == false ){
				intensity_changed = true;
				intensity += intensity_inc;
				if(intensity < 0)
					intensity = 0.0f;
				else if(intensity > 1.0)
					intensity = 1.0;

				skyStateSet->getUniform("alpha")->set(float(intensity/max_intensity));

				float w_intensity = intensity * max_intensity;
				
				if(weather == 1 || weather == 3){
					precipitationEffect->rain(w_intensity);
					frict = max_frict - (w_intensity/max_intensity)*rain_coef;
				}
				else{
					precipitationEffect->snow(w_intensity);
					frict = max_frict - (w_intensity/max_intensity)*snow_coef;
				}

				for(int i = 0; i < (int)players.size(); i++){
					players.at(i)->car->setFriction(0,frict);
					players.at(i)->car->setFriction(1,frict);
					players.at(i)->car->setFriction(2,frict);
					players.at(i)->car->setFriction(3,frict);
				}
			}
			else if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % 2 > 0 && intensity_changed == true ){
				intensity_changed = false;
			}
		}
		// dynamic sky changes
		else if( weather == 0 ){
			if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % weather_change_time == 0 && weather_changed == false ){
				weather_changed = true;
				srand(now_tick);
				float final_intensity = (rand()%(int)11)/(11.0);
				intensity_inc = (final_intensity - intensity)/float(weather_change_time/2.0);
			}
			else if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % weather_change_time > 0 && weather_changed == true ){
				weather_changed = false;
				srand(now_tick);
				weather_change_time = rand()%60+60;
			}

			if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % 2 == 0 && intensity_changed == false ){
				intensity_changed = true;
				intensity += intensity_inc;
				if(intensity < 0)
					intensity = 0.0f;
				else if(intensity > 0.3)
					intensity = 0.3;

				skyStateSet->getUniform("alpha")->set(float(intensity/max_intensity));
			}
			else if( (int)osg::Timer::instance()->delta_s(weather_timer, now_tick) % 2 > 0 && intensity_changed == true ){
				intensity_changed = false;
			}
		}

/*		if(keyEH->IsKeyPressed('p')){
			pause = (pause)?false:true;
		}*/	

		if(pause){
			viewer->frame();
			continue;
		}	


		// computing tyres temperatures - undone!

		//float speed = fabs(players.at(0)->car->getSpeed());

		//if( speed > 5){
		//	for(int i = 0; i < 4; i++){
		//		if(players.at(0)car->vehicle->getWheelInfo(i).m_skidInfo == 1.0){
		//			if(temperature[i] < 1.0){
		//				temperature[i] += dt*0.002;
		//			}
		//			else if(temperature[i] > 1.0){
		//				if(temperature[i] > 1.6)
		//					temperature[i] -= dt*0.15;
		//				else if(temperature[i] > 1.4)
		//					temperature[i] -= dt*0.09;
		//				else if(temperature[i] > 1.2)
		//					temperature[i] -= dt*0.04;
		//				else
		//					temperature[i] -= dt*0.01;	
		//			}
		//		}
		//		else{
		//			if(speed < 30){
		//				temperature[i] += (1 - players.at(0)->car->vehicle->getWheelInfo(i).m_skidInfo) * dt * 0.002;
		//			}
		//			else if(speed < 70){
		//				temperature[i] += (1 - players.at(0)->car->vehicle->getWheelInfo(i).m_skidInfo) * dt * 0.007;
		//			}
		//			else if(speed < 150){
		//				temperature[i] += (1 - player->car->vehicle->getWheelInfo(i).m_skidInfo) * dt * 0.02;
		//			}
		//			else{
		//				temperature[i] += (1 - player->car->vehicle->getWheelInfo(i).m_skidInfo) * dt * 0.05;
		//			}
		//		}
		//	}
		//}

		//
		//for(int i = 0; i < 4; i++){
		//	temperature[i] -= dt*0.0005;

		//	if(temperature[i] > 2.0)
		//		temperature[i] = 2.0;
		//	else if(temperature[i] < 0.8)
		//		temperature[i] = 0.8;
		//}
		
		///////////////
		// CAR UPDATE
		///////////////
			
		// simulate physics
		physicalWorld->stepSimulation(dt);
		physicalWorld->updateAabbs();

		dt = osg::Timer::instance()->delta_m(frame_tick, now_tick);
		for(int i = 0; i < (int)players.size(); i++){
			if(players.size() == 1)
				players.at(i)->car->updateVehicle(frame_tick, dt, Vehicle::UPDATE_ALL | Vehicle::EXPERIMENTAL_ANALOG);
			else
				players.at(i)->car->updateVehicle(frame_tick, dt, Vehicle::UPDATE_ALL);
			
			// update HUDs
			players.at(i)->hud->updateTachometer(players.at(i)->car->getHudData());
			if( i == 1 ){
				players.at(0)->hud->updateMap(players.at(1)->car->getHudData().pos);
				players.at(1)->hud->updateMap(players.at(0)->car->getHudData().pos);	
			}
		}

		///////////////
		// CAR UPDATE END
		///////////////


		// getting times in sectors
		for(int i = 0; i < (int)players.size(); i++){
			char txt[256];

			osg::Vec3d newPos = players.at(i)->car->getChassisPAT()->getPosition();

			// game init
			if( players.at(i)->sector1 == 0 && players.at(i)->sector2 == 0 && players.at(i)->sector3 == 0){
				// car crosses the starting line for the first time
				if( players.at(i)->car->getChassisPAT()->getPosition().y() >= 0 && ((2*3.14*0.338)*((players.at(i)->car->vehicle->m_wheelInfo.at(2).m_rotation * 57.296)/360.0) > 5)){
					// store time of incoming to the sector 1
					players.at(i)->sector1 = 1;
					players.at(i)->lap_time = now_tick;
					players.at(i)->lap++;
					char lap_str[4];
					sprintf(lap_str,"%d",players.at(i)->lap);
					players.at(i)->hud->HUDLapCnt->setText(lap_str);
				}	
			}
			// game is running
			else{
				// compute distance between car center and sectors lines
				double d;
				double xa = -15.0f, ya = 0.0f, xb = 15.0f, yb = 0.0f;
				double xa2 = 784.0f, ya2 = -311.0f, xb2 = 826.0f, yb2 = 279.0f;
				double xa3 = 362.0f, ya3 = -1034.0f, xb3 = 417.0f, yb3 = -1040.0f;

				// we are in sector1 -> compute distance to end of sector 1
				if( players.at(i)->sector2 == 0 && players.at(i)->sector1 != 0){
					double cross_p2 = ((yb2-ya2)*(players.at(i)->car->getChassisPAT()->getPosition().x()-xa2)+(xb2-xa2)*(players.at(i)->car->getChassisPAT()->getPosition().y()-ya2))*((yb2-ya2)*(players.at(i)->car->getChassisPAT()->getPosition().x()-xa2)+(xb2-xa2)*(players.at(i)->car->getChassisPAT()->getPosition().y()-ya2));
					d = sqrt((cross_p2)/((xb2-xa2)*(xb2-xa2) + (yb2-ya2)*(yb2-ya2)));
					//sprintf(txt,"sector1\n%.3f",d);
					// car crossed the line and is now in sector2
					if( players.at(i)->d_old < d && players.at(i)->car->getChassisPAT()->getPosition().x() >= xa2-10 && players.at(i)->car->getChassisPAT()->getPosition().y() >= ya2-10){
						// store time for sector1
						players.at(i)->sector1 = osg::Timer::instance()->delta_s(players.at(i)->lap_time, now_tick);
						// time continues in sector2
						players.at(i)->sector2 = 1;
						//cout<<players.at(i)->sector1<<" ";

						if( players.at(i)->bestS1 > 0 ){
							int mins;
							float secs;

							mins = fabs(players.at(i)->sector1-players.at(i)->bestS1)/60;
							secs = fabs(players.at(i)->sector1-players.at(i)->bestS1) - 60*mins;
							if(players.at(i)->sector1 < players.at(i)->bestS1){
								sprintf(txt,"-%d:%06.3f",mins,secs);
								players.at(i)->hud->setSectorTime(txt,10000,false);
							}
							else{
								sprintf(txt,"+%d:%06.3f",mins,secs);
								players.at(i)->hud->setSectorTime(txt,10000,true);
							}

							mins = players.at(i)->bestS2/60;
							secs = players.at(i)->bestS2 - 60*mins;						
							sprintf(txt,"s2: %d:%06.3f",mins,secs);
							players.at(i)->hud->HUDBestLapS1->setText(txt);
						}

						if( players.at(i)->recordS1 > 0 ){
							int mins;
							float secs;

							mins = fabs(players.at(i)->sector1-players.at(i)->recordS1)/60;
							secs = fabs(players.at(i)->sector1-players.at(i)->recordS1) - 60*mins;
							if(players.at(i)->sector1 < players.at(i)->recordS1){
								sprintf(txt,"-%d:%06.3f",mins,secs);
								players.at(i)->hud->setRecordSectorTime(txt,10000,false);
							}
							else{
								sprintf(txt,"+%d:%06.3f",mins,secs);
								players.at(i)->hud->setRecordSectorTime(txt,10000,true);
							}
						}

						// load track record
						if( mysql_connected == true && players.at(i)->online == true && slow_connection == 0 ){
							char query[1024];
							sql::ResultSet *res;
							sprintf(query,"SELECT `lap_time`,`name`,`sector1`,`sector2` FROM `c20348dip`.`times` WHERE `car_type` = '%d' AND `weather` = '%d-%d'  ORDER BY `lap_time` ASC LIMIT 1", cars_used[i], weather, intense);
							res = stmt->executeQuery(query);

							while (res->next()){
								char lap_text[64],lapS1_text[64];
								int mins;
								float secs;

								mins = res->getDouble("lap_time")/60;
								secs = res->getDouble("lap_time") - 60*mins;
								sprintf(lap_text,"%d:%06.3f",mins,secs);
									
								mins = res->getDouble("sector2")/60;
								secs = res->getDouble("sector2") - 60*mins;
								sprintf(lapS1_text,"s2: %d:%06.3f",mins,secs);

								for(int ii = 0; ii < (int)players.size(); ii++){
									players.at(ii)->recordLapT = res->getDouble("lap_time");
									players.at(ii)->recordS1 = res->getDouble("sector1");
									players.at(ii)->recordS2 = res->getDouble("sector2");
									players.at(ii)->hud->HUDRecord->setText(lap_text);
									players.at(ii)->hud->HUDRecordman->setText(res->getString("name"));
									players.at(ii)->hud->HUDRecordS1->setText(lapS1_text);
								}
							}

							delete res;
						}
						else if( mysql_connected == true && players.at(i)->online == true ){
							char lapS1_text[64];
							int mins;
							float secs;

							mins = players.at(i)->recordS2/60;
							secs = players.at(i)->recordS2 - 60*mins;
							sprintf(lapS1_text,"s2 %d:%06.3f",mins,secs);

							players.at(i)->hud->HUDRecordS1->setText(lapS1_text);
						}

						players.at(i)->d_old = 10000;
					}
					else
						players.at(i)->d_old = d;
				}
				// we're in sector2
				else if( players.at(i)->sector3 == 0 && players.at(i)->sector2 != 0){
					double cross_p3 = ((yb3-ya3)*(players.at(i)->car->getChassisPAT()->getPosition().x()-xa3)+(xb3-xa3)*(players.at(i)->car->getChassisPAT()->getPosition().y()-ya3))*((yb3-ya3)*(players.at(i)->car->getChassisPAT()->getPosition().x()-xa3)+(xb3-xa3)*(players.at(i)->car->getChassisPAT()->getPosition().y()-ya3));	
					d = sqrt((cross_p3)/((xb3-xa3)*(xb3-xa3) + (yb3-ya3)*(yb3-ya3)));
					//sprintf(txt,"sector2\n%.3f",d);
					// car crossed the line and is now in sector3
					if( players.at(i)->d_old < d && players.at(i)->car->getChassisPAT()->getPosition().x() <= xb3+10 && players.at(i)->car->getChassisPAT()->getPosition().y() <= ya3+10){
						// store time for sector2
						players.at(i)->sector2 = osg::Timer::instance()->delta_s(players.at(i)->lap_time, now_tick);
						// time continues in sector3
						players.at(i)->sector3 = 1;
						//cout<<players.at(i)->sector2<<" ";

						if( players.at(i)->bestS2 > 0 ){
							int mins;
							float secs;

							mins = fabs(players.at(i)->sector2-players.at(i)->bestS2)/60;
							secs = fabs(players.at(i)->sector2-players.at(i)->bestS2) - 60*mins;
							if(players.at(i)->sector2 < players.at(i)->bestS2){
								sprintf(txt,"-%d:%06.3f",mins,secs);
								players.at(i)->hud->setSectorTime(txt,10000,false);
							}
							else{
								sprintf(txt,"+%d:%06.3f",mins,secs);
								players.at(i)->hud->setSectorTime(txt,10000,true);
							}

							mins = players.at(i)->bestLapT/60;
							secs = players.at(i)->bestLapT - 60*mins;						
							sprintf(txt,"s3: %d:%06.3f",mins,secs);
							players.at(i)->hud->HUDBestLapS1->setText(txt);
						}

						if( players.at(i)->recordS2 > 0 ){
							int mins;
							float secs;

							mins = fabs(players.at(i)->sector2-players.at(i)->recordS2)/60;
							secs = fabs(players.at(i)->sector2-players.at(i)->recordS2) - 60*mins;
							if(players.at(i)->sector2 < players.at(i)->recordS2){
								sprintf(txt,"-%d:%06.3f",mins,secs);
								players.at(i)->hud->setRecordSectorTime(txt,10000,false);
							}
							else{
								sprintf(txt,"+%d:%06.3f",mins,secs);
								players.at(i)->hud->setRecordSectorTime(txt,10000,true);
							}
						}

						// load track record
						if( mysql_connected == true && players.at(i)->online == true && slow_connection == 0 ){
							char query[1024];
							sql::ResultSet *res;
							sprintf(query,"SELECT `lap_time`,`name`,`sector1`,`sector2` FROM `c20348dip`.`times` WHERE `car_type` = '%d' AND `weather` = '%d-%d'  ORDER BY `lap_time` ASC LIMIT 1", cars_used[i], weather, intense);
							res = stmt->executeQuery(query);

							while (res->next()){
								char lap_text[64],lapS1_text[64];
								int mins;
								float secs;

								mins = res->getDouble("lap_time")/60;
								secs = res->getDouble("lap_time") - 60*mins;
								sprintf(lap_text,"%d:%06.3f",mins,secs);
									
								sprintf(lapS1_text,"s3: %d:%06.3f",mins,secs);

								for(int ii = 0; ii < (int)players.size(); ii++){
									players.at(ii)->recordLapT = res->getDouble("lap_time");
									players.at(ii)->recordS1 = res->getDouble("sector1");
									players.at(ii)->recordS2 = res->getDouble("sector2");
									players.at(ii)->hud->HUDRecord->setText(lap_text);
									players.at(ii)->hud->HUDRecordman->setText(res->getString("name"));
									players.at(ii)->hud->HUDRecordS1->setText(lapS1_text);
								}
							}
							delete res;
						}
						else if( mysql_connected == true && players.at(i)->online == true ){
							char lapS1_text[64];
							int mins;
							float secs;

							mins = players.at(i)->recordLapT/60;
							secs = players.at(i)->recordLapT - 60*mins;
							sprintf(lapS1_text,"s3: %d:%06.3f",mins,secs);

							players.at(i)->hud->HUDRecordS1->setText(lapS1_text);
						}

						players.at(i)->d_old = 10000;
					}
					else
						players.at(i)->d_old = d;
				}
				// we're in final sector
				else if( players.at(i)->sector3 != 0 && players.at(i)->sector1 != 0 ){
					double cross_p = ((yb-ya)*(players.at(i)->car->getChassisPAT()->getPosition().x()-xa)+(xb-xa)*(players.at(i)->car->getChassisPAT()->getPosition().y()-ya))*((yb-ya)*(players.at(i)->car->getChassisPAT()->getPosition().x()-xa)+(xb-xa)*(players.at(i)->car->getChassisPAT()->getPosition().y()-ya));
					d = sqrt((cross_p)/((xb-xa)*(xb-xa) + (yb-ya)*(yb-ya)));
					//sprintf(txt,"sector3\n%.3f",d);
					// car crossed the finnish line and is now in sector1
					if( players.at(i)->d_old < d && players.at(i)->car->getChassisPAT()->getPosition().x() <= xb+10){
						// store time for the whole lap
						players.at(i)->lapT = players.at(i)->sector3 + osg::Timer::instance()->delta_s(players.at(i)->lap_time, now_tick);
						players.at(i)->lap_time = now_tick;
						players.at(i)->d_old = 10000;
						//cout<<players.at(i)->lapT<<endl;
						players.at(i)->lap++;
						char lap_str[4];
						sprintf(lap_str,"%d",players.at(i)->lap);
						players.at(i)->hud->HUDLapCnt->setText(lap_str);

						if( players.at(i)->bestLapT > 0 ){
							int mins;
							float secs;

							mins = fabs(players.at(i)->lapT-players.at(i)->bestLapT)/60;
							secs = fabs(players.at(i)->lapT-players.at(i)->bestLapT) - 60*mins;
							if(players.at(i)->lapT < players.at(i)->bestLapT){
								sprintf(txt,"-%d:%06.3f",mins,secs);
								players.at(i)->hud->setSectorTime(txt,10000,false);
							}
							else{
								sprintf(txt,"+%d:%06.3f",mins,secs);
								players.at(i)->hud->setSectorTime(txt,10000,true);
							}

							mins = players.at(i)->bestS1/60;
							secs = players.at(i)->bestS1 - 60*mins;						
							sprintf(txt,"s1: %d:%06.3f",mins,secs);
							players.at(i)->hud->HUDBestLapS1->setText(txt);
						}

						if( players.at(i)->recordLapT > 0 ){
							int mins;
							float secs;

							mins = fabs(players.at(i)->lapT-players.at(i)->recordLapT)/60;
							secs = fabs(players.at(i)->lapT-players.at(i)->recordLapT) - 60*mins;
							if(players.at(i)->lapT < players.at(i)->recordLapT){
								sprintf(txt,"-%d:%06.3f",mins,secs);
								players.at(i)->hud->setRecordSectorTime(txt,10000,false);
							}
							else{
								sprintf(txt,"+%d:%06.3f",mins,secs);
								players.at(i)->hud->setRecordSectorTime(txt,10000,true);
							}
						}

						// show last lap time
						char lap_text[64];
						int mins;
						float secs;
						mins = players.at(i)->lapT/60;
						secs = players.at(i)->lapT - 60*mins;
						sprintf(lap_text,"%d:%06.3f",mins,secs);
						players.at(i)->hud->HUDLastLap->setText(lap_text);
						
						// compare with best lap
						if( players.at(i)->bestLapT > players.at(i)->lapT || players.at(i)->bestLapT == -1){
							float old_best = players.at(i)->bestLapT;
							players.at(i)->hud->HUDBestLap->setText(lap_text);
							mins = players.at(i)->sector1/60;
							secs = players.at(i)->sector1 - 60*mins;
							sprintf(lap_text,"s1: %d:%06.3f",mins,secs);
							players.at(i)->hud->HUDBestLapS1->setText(lap_text);
							players.at(i)->bestS1 = players.at(i)->sector1;
							players.at(i)->bestS2 = players.at(i)->sector2;
							players.at(i)->bestLapT = players.at(i)->lapT;

							if(mysql_connected == true && players.at(i)->online == true ){
								char query[1024];
								if( old_best == -1 )
									sprintf(query, "INSERT INTO `c20348dip`.`times` (`id` ,`name` ,`sector1` ,`sector2` ,`lap_time`, `car_type`, `weather` )VALUES (NULL , '%s', '%.3f', '%.3f', '%.3f', '%d', '%d-%d');",players.at(i)->name.c_str(),players.at(i)->sector1,players.at(i)->sector2,players.at(i)->lapT, cars_used[i], weather, intense);
								else
									sprintf(query, "UPDATE `c20348dip`.`times` SET `sector1` = '%.3f', `sector2` = '%.3f', `lap_time` = '%.3f' WHERE `name` = '%s' AND `car_type` = '%d' AND `weather` = '%d-%d' LIMIT 1;",players.at(i)->sector1,players.at(i)->sector2,players.at(i)->lapT,players.at(i)->name.c_str(), cars_used[i], weather, intense);
								stmt->execute(query);
							}

						}

						// load track record
						if( mysql_connected == true && players.at(i)->online == true && slow_connection == 0 ){
							char query[1024];
							sql::ResultSet *res;
							sprintf(query,"SELECT `lap_time`,`name`,`sector1`,`sector2` FROM `c20348dip`.`times` WHERE `car_type` = '%d' AND `weather` = '%d-%d'  ORDER BY `lap_time` ASC LIMIT 1", cars_used[i], weather, intense);
							res = stmt->executeQuery(query);

							while (res->next()){
								char lap_text[64],lapS1_text[64];
								int mins;
								float secs;

								mins = res->getDouble("lap_time")/60;
								secs = res->getDouble("lap_time") - 60*mins;
								sprintf(lap_text,"%d:%06.3f",mins,secs);
									
								mins = res->getDouble("sector1")/60;
								secs = res->getDouble("sector1") - 60*mins;
								sprintf(lapS1_text,"s1: %d:%06.3f",mins,secs);

								for(int ii = 0; ii < (int)players.size(); ii++){
									players.at(ii)->recordLapT = res->getDouble("lap_time");
									players.at(ii)->recordS1 = res->getDouble("sector1");
									players.at(ii)->recordS2 = res->getDouble("sector2");
									players.at(ii)->hud->HUDRecord->setText(lap_text);
									players.at(ii)->hud->HUDRecordman->setText(res->getString("name"));
									players.at(ii)->hud->HUDRecordS1->setText(lapS1_text);
								}
							}
							delete res;
						}
						else if( mysql_connected == true && players.at(i)->online == true ){
							char lapS1_text[64];
							int mins;
							float secs;

							mins = players.at(i)->recordS1/60;
							secs = players.at(i)->recordS1 - 60*mins;
							sprintf(lapS1_text,"s1: %d:%06.3f",mins,secs);

							players.at(i)->hud->HUDRecordS1->setText(lapS1_text);
						}

						// erase times
						players.at(i)->sector3 = 0;
						players.at(i)->sector2 = 0;
						// start measuring time for new lap
						players.at(i)->sector1 = 1;
					}
					else
						players.at(i)->d_old = d;
				}

				// show actual lap time
				char lap_text[64];
				int mins;
				float secs;
				double dur = osg::Timer::instance()->delta_s(players.at(i)->lap_time, now_tick);
				mins = dur/60;
				secs = dur - 60*mins;
				sprintf(lap_text,"%02d:%06.3f",mins,secs);
				players.at(i)->hud->HUDactTime->setText(lap_text);
			}


			// dynamic change of camera
			osg::Matrixd m;
			m.identity();
			double offset = fabs(players.at(i)->car->getSpeed())/30.0f;
			if(offset > 7)
				offset = 7;
			m.setTrans(0,-2,-10.0*players.size()-offset);
			m.setRotate(osg::Quat(3.14/2.1,osg::Vec3f(-1,0,0)));
			players.at(i)->fc->changeOffset(m);
		}

        // render
        viewer->frame();
		frame_tick = now_tick;
    }

	if( mysql_connected == true && slow_connection == 1 ){
		for(int i = 0; i < (int)players.size(); i++){
			if( players.at(i)->online == true ){
				char query[1024];
				if( players.at(i)->bestLapTmp > -1 && players.at(i)->recordLapT == -1 )
					sprintf(query, "INSERT INTO `c20348dip`.`times` (`id` ,`name` ,`sector1` ,`sector2` ,`lap_time`, `car_type`, `weather` )VALUES (NULL , '%s', '%.3f', '%.3f', '%.3f', '%d', '%d-%d');",players.at(i)->name.c_str(),players.at(i)->bestS1,players.at(i)->bestS2,players.at(i)->bestLapT, cars_used[i], weather, intense);
				else if( players.at(i)->bestLapTmp > players.at(i)->bestLapT )
					sprintf(query, "UPDATE `c20348dip`.`times` SET `sector1` = '%.3f', `sector2` = '%.3f', `lap_time` = '%.3f' WHERE `name` = '%s' AND `car_type` = '%d' AND `weather` = '%d-%d' LIMIT 1;",players.at(i)->bestS1,players.at(i)->bestS2,players.at(i)->bestLapT,players.at(i)->name.c_str(), cars_used[i], weather, intense);
				stmt->execute(query);
			}
		}
	}


	if(connect_to_mysql == true){
		delete con;
		if(mysql_connected == true){
			delete stmt;
		}
	}
	
	// remove settings file
	remove( "data/Brno/tmp" );

    return 0;
}

