/*
	Physical Simulation in Graphics Scene - master's thesis
    Copyright (C) 2010-2011  Marin Javorka, FIT VUT in Brno, http://www.fit.vutbr.cz, marian.javorka@deq.sk

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/
#include "physics.h"

Physics *physics = new Physics();

Physics::Physics(){

	g = ( float )9.81;
	gravity = btVector3(0, 0, (-1*g) );

	temperature = 293.15; //K		
	
	p = 101325;// Pa
	R_spec = 287.058;// J/(kgK)

	airD = p/(temperature*R_spec);

	return;
};

// init physics
btDynamicsWorld* Physics::initPhysics(){

	// default configuration
	btDefaultCollisionConfiguration* collisionConfig = new btDefaultCollisionConfiguration();

	// colision dispatcher
	btCollisionDispatcher* dispatcher = new btCollisionDispatcher( collisionConfig );

	// solver
	btConstraintSolver* solver = new btSequentialImpulseConstraintSolver;

	// set-up world min and max
	btVector3 worldAabbMin( -20000, -20000, -20000 );
	btVector3 worldAabbMax( 20000, 20000, 20000 );

	// set-up interface
	unsigned short maxHandles = 2000;
	btBroadphaseInterface* interfce = new btAxisSweep3( worldAabbMin, worldAabbMax, maxHandles);

	// create dynamic world
	btDynamicsWorld* dynamicWorld = new btDiscreteDynamicsWorld( dispatcher, interfce, solver, collisionConfig );

	// set gravity
	dynamicWorld->setGravity( gravity );

	return dynamicWorld;
};

// create rigid body and add it to the world
btRigidBody* Physics::createRigidBody(btDynamicsWorld* world, float mass, const btTransform& startTransform, btCollisionShape* shape, double inerX, double inerY, double inerZ){
	
	bool isDynamic = true;
	// if mass is 0, rigid body is static, otherwise dynamic
	if(mass == 0.0f)
		isDynamic = false;

	// set local inertia
	btVector3 localInertia(inerX, inerY, inerZ);

	// calculate inertia if rigid body is dynamic
	if(isDynamic)
		shape->calculateLocalInertia(mass, localInertia);

	// using motion state
	btDefaultMotionState* motionState = new btDefaultMotionState(startTransform);

	// create body
	btRigidBody* body = new btRigidBody(mass, motionState, shape, localInertia);

	// add body to world
	world->addRigidBody(body);

	return body;
};

// create rigid body
btRigidBody* Physics::createRigidBody(float mass, const btTransform& startTransform, btCollisionShape* shape){
	
	bool isDynamic = true;
	// if mass is 0, rigid body is static, otherwise dynamic
	if(mass == 0.0f)
		isDynamic = false;

	// set local inertia
	btVector3 localInertia(0, 0, 0);

	// calculate inertia if rigid body is dynamic
	if(isDynamic)
		shape->calculateLocalInertia(mass, localInertia);

	// using motion state
	btDefaultMotionState* motionState = new btDefaultMotionState(startTransform);

	// create body
	btRigidBody* body = new btRigidBody(mass, motionState, shape, localInertia);

	return body;
};

btCollisionShape* Physics::createCollisionShapeFromBvhTriangleMeshShape(const std::string file){
	string str;						// line in string
	string::size_type comment;		// position of comment char #
	char line[256];					// line in char[]
	ifstream shapeFile(file.c_str());		// konfiguracny subor
	int inPoints = 0;				// flag, ci nacitavame points tabulku
	int inTriangles = 0;			// flag, ci nacitavame triangles tabulku
	vector<osg::Vec3f> points;		// vektor bodov

	btTriangleMesh* trimesh = new btTriangleMesh();

	// nacitavanie celeho suboru po riadokch
	while( shapeFile.getline(line, 256) ){
		str.assign(line);

		// zistenie pozicie komentara
		comment = str.find_first_of('#');
    
		// ak sa komentar nasiel,tak sa vsetko za nim odstrani
		if(comment < 256)
		str.erase(comment);
    
		// ak mame neprazdny string, tak naplnime strukturu dnsConfig
		if(str.length()>0){
			// test na points tabulku
			if( str.find( "points", 0 ) != string::npos ){
				inTriangles = 0;
				inPoints = 1;
				continue;
			}
			// test na vertices tabulku
			if( str.find( "triangles", 0 ) != string::npos ){
				inTriangles = 1;
				inPoints = 0;
				continue;
			}

			if(inPoints == 1){
				osg::Vec3f point;
				if(sscanf(str.c_str(), "%f %f %f",&point.x(), &point.y(), &point.z()) < 3){
					// jedna sa o {,}, alebo nejaky bordel...tak otestujeme ci je }, ak ano,tak sme vonku z points tabulky
					if( str.find( "}", 0 ) != string::npos )
						inPoints = 0;
				}
				else{
					points.push_back(point);	
				}
			}

			if(inTriangles == 1){
				int x,y,z;
				if(sscanf(str.c_str(), "%d %d %d",&x, &y, &z) < 3){
					// jedna sa o {,}, alebo nejaky bordel...tak otestujeme ci je }, ak ano,tak sme vonku z vertices tabulky
					if( str.find( "}", 0 ) != string::npos )
						inTriangles = 0;
				}
				else{
					trimesh->addTriangle(btVector3(points.at(x).x(), points.at(x).z(), points.at(x).y()),
										btVector3(points.at(y).x(), points.at(y).z(), points.at(y).y()),
										btVector3(points.at(z).x(), points.at(z).z(), points.at(z).y()));
				}
			}
		}
	}

	btCollisionShape* terrain = new btBvhTriangleMeshShape(trimesh,true);
	
	return terrain;
};

btConvexHullShape* Physics::createConvexHullShapeFromConvexTriangleMeshShape(const std::string file){
	string str;						// line in string
	string::size_type comment;		// position of comment char #
	char line[256];					// line in char[]
	ifstream shapeFile(file.c_str());		// konfiguracny subor
	int inPoints = 0;				// flag, ci nacitavame points tabulku
	int inTriangles = 0;			// flag, ci nacitavame triangles tabulku
	vector<osg::Vec3f> points;		// vektor bodov
	btConvexHullShape* chassisShapePr = new btConvexHullShape();

	btTriangleMesh* trimesh = new btTriangleMesh();

	// nacitavanie celeho suboru po riadokch
	while( shapeFile.getline(line, 256) ){
		str.assign(line);

		// zistenie pozicie komentara
		comment = str.find_first_of('#');
    
		// ak sa komentar nasiel,tak sa vsetko za nim odstrani
		if(comment < 256)
		str.erase(comment);
    
		// ak mame neprazdny string, tak naplnime strukturu dnsConfig
		if(str.length()>0){
			// test na points tabulku
			if( str.find( "points", 0 ) != string::npos ){
				inTriangles = 0;
				inPoints = 1;
				continue;
			}
			// test na vertices tabulku
			if( str.find( "triangles", 0 ) != string::npos ){
				inTriangles = 1;
				inPoints = 0;
				continue;
			}

			if(inPoints == 1){
				osg::Vec3f point;
				if(sscanf(str.c_str(), "%f %f %f",&point.x(), &point.y(), &point.z()) < 3){
					// jedna sa o {,}, alebo nejaky bordel...tak otestujeme ci je }, ak ano,tak sme vonku z points tabulky
					if( str.find( "}", 0 ) != string::npos )
						inPoints = 0;
				}
				else{
					points.push_back(point);	
				}
			}

			if(inTriangles == 1){
				int x,y,z;
				if(sscanf(str.c_str(), "%d %d %d",&x, &y, &z) < 3){
					// jedna sa o {,}, alebo nejaky bordel...tak otestujeme ci je }, ak ano,tak sme vonku z vertices tabulky
					if( str.find( "}", 0 ) != string::npos )
						inTriangles = 0;
				}
				else{
					trimesh->addTriangle(btVector3(points.at(x).x(), points.at(x).z(), points.at(x).y()),
										btVector3(points.at(y).x(), points.at(y).z(), points.at(y).y()),
										btVector3(points.at(z).x(), points.at(z).z(), points.at(z).y()));
				}
			}
		}
	}

	btConvexShape* tmpConvexShape = new btConvexTriangleMeshShape(trimesh);

	//create a hull approximation
	btShapeHull* hull = new btShapeHull(tmpConvexShape);
	btScalar margin = tmpConvexShape->getMargin();
	hull->buildHull(margin);
	tmpConvexShape->setUserPointer(hull);
	
	for (int i=0;i<hull->numVertices();i++)
	{
		chassisShapePr->addPoint(hull->getVertexPointer()[i]);
	}

	delete tmpConvexShape;
	delete hull;

	return chassisShapePr;
};