/*
	Physical Simulation in Graphics Scene - master's thesis
    Copyright (C) 2010-2011  Marin Javorka, FIT VUT in Brno, http://www.fit.vutbr.cz, marian.javorka@deq.sk

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/
#include "vehicle.h"

osg::Node* Vehicle::findNodeByName(osg::Node* node,const std::string& name){

	// je to ten isty
	if(node->getName() == name)
		return node;

	osg::Group* group = node->asGroup();
	if(group)
		for(unsigned int i = 0; i<group->getNumChildren(); i++){
			osg::Node* found = findNodeByName(group->getChild(i),name);
			if(found)
				return found;
		}

	return 0;
}

Vehicle::Vehicle(btDynamicsWorld* m_world, osg::Vec3f pos, VehicleModel model, KeyMap keyMap, bool get_manual){
	chassisPAT = new osg::PositionAttitudeTransform;

	vehicleModel = model;

	// add model to PAT
	chassisPAT->addChild(vehicleModel.chassisModel.get());

	handler = keyEH;

	throttle = 0;
	brake = 0;

	handbrake = false;

	gear = 1;

	manual = get_manual;

	air_tunnel = 1.0;

	wheelDirectionCS0 = btVector3(0,0,-1);						// direction of ray cast (to the ground)
	wheelAxleCS = btVector3(1,0,0);								// axle of rotation (X)

	gEngineForce = 0.f;
	gBreakingForce = 0.f;
	gVehicleSteering = 0.f;
	//maxEngineForce = 0.f;//this should be engine/velocity dependent

//	maxBreakingForce = 3000.f;

	steeringIncrement = 0.01f;
	steeringClamp = 0.8f;

	vehicle = 0;

	driving = ACKERMAN;

	//chassisLength = 4.49f;
	//chassisWidth = 1.92f;
	//chassisHeight = 1.108f;

	//mass = 1400.f;

	world = m_world;

	resetPosition = pos;

	keys = keyMap;

	gear_ratio.push_back(0);	// reverse
	gear_ratio.push_back(0);	// neutral

	engine_rpm = 0;
	tachometer_rpm = 0;
	clutch = 1;
};

void Vehicle::configVehicle(const char* configFile){
	// set-up chassis collision shape
	//chassisShape = new btConvexHullShape();
	//createCollisionShape(chassisShape, "data/Dodge Viper GTS_collision.x");

	parseConfig(configFile);

	btCompoundShape* compound = new btCompoundShape();
	btTransform localTrans;
	localTrans.setIdentity();
	localTrans.setOrigin(btVector3(0,0,0.07));
	//localTrans.setOrigin(btVector3(0,0,1.0));

	compound->addChildShape(localTrans, vehicleModel.chassisShape);

	tr.setOrigin(btVector3(0.f,0.f,0.f));

	chassis = physics->createRigidBody(world, mass, tr, compound);

	chassis->setUserPointer(vehicleModel.chassisModel.get());
	chassisPAT->setUpdateCallback(new RigidBodyUpdateCallback(chassis));

	resetVehicle();

	createVehicle();

	createWheel(frontWheelChassisConnectionPoint, true, 0);
	createWheel(btVector3(-frontWheelChassisConnectionPoint.x(), frontWheelChassisConnectionPoint.y(), frontWheelChassisConnectionPoint.z()), true, 1);
	createWheel(btVector3(-rearWheelChassisConnectionPoint.x(), rearWheelChassisConnectionPoint.y(), rearWheelChassisConnectionPoint.z()), false, 2);
	createWheel(rearWheelChassisConnectionPoint, false, 3);

	osg::StateSet* stateSet = new osg::StateSet();
	//stateSet->setTextureAttributeAndModes(0, texture, osg::StateAttribute::ON);

	// turn on blending
	stateSet->setMode(GL_BLEND, osg::StateAttribute::ON);

	// disable depth_test
	//stateSet->setMode(GL_DEPTH_TEST, osg::StateAttribute::OFF);
	stateSet->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	// set StateSet
	vehicleModel.chassisModel.get()->setStateSet(stateSet);

	findNodeByName(vehicleModel.chassisModel.get(),"brzda_on")->setNodeMask(0x0);
	findNodeByName(vehicleModel.chassisModel.get(),"rev_light")->setNodeMask(0x0);

	osg::StateSet* stateSetNoLight = new osg::StateSet();

	// set StateSet
    stateSetNoLight->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE|osg::StateAttribute::OFF);
	findNodeByName(vehicleModel.chassisModel.get(),"brzda_on")->setStateSet(stateSetNoLight);
	findNodeByName(vehicleModel.chassisModel.get(),"rev_light")->setStateSet(stateSetNoLight);
};

void Vehicle::resetVehicle(){
	gVehicleSteering = 0.f;
	gear = 1;
	engine_rpm = lowest_rpm;
	tachometer_rpm = 0;
	clutch = 1;
	btTransform chassisCOM = btTransform::getIdentity();
	chassisCOM.setOrigin(btVector3(resetPosition.x(),resetPosition.y(),resetPosition.z()));
	chassis->setCenterOfMassTransform(chassisCOM);
	chassis->setLinearVelocity(btVector3(0,0,0));
	chassis->setAngularVelocity(btVector3(0,0,0));
	world->getBroadphase()->getOverlappingPairCache()->cleanProxyFromPairs(chassis->getBroadphaseHandle(), world->getDispatcher());

	if(vehicle){
		vehicle->resetSuspension();
		for(int i=0; i<vehicle->getNumWheels(); i++)
			vehicle->updateWheelTransform(i,true);
	}
};

void Vehicle::createVehicle(){
	rayCaster = new btDefaultVehicleRaycaster(world);
	tuning = btRaycastVehicle::btVehicleTuning();
	vehicle = new btRaycastVehicle(tuning, chassis, rayCaster);

	// vehicle activation
	chassis->setActivationState(DISABLE_DEACTIVATION);

	world->addVehicle(vehicle);

	// set coordinate system
	vehicle->setCoordinateSystem(0,2,1);	
};

void Vehicle::createWheel(const btVector3 &connectionPointCS0, bool isFrontWheel, int wheelID){
	if(isFrontWheel){

		vehicle->addWheel(connectionPointCS0, wheelDirectionCS0, wheelAxleCS, frontSuspensionRestLength, frontWheelRadius, tuning, isFrontWheel);

		wheel[wheelID] = physics->createRigidBody(world, 20, tr, vehicleModel.wheelShape[wheelID]);
		wheel[wheelID]->setUserPointer(vehicleModel.wheelModel[wheelID].get());
		wheelPAT[wheelID] = new osg::PositionAttitudeTransform;

		wheelPAT[wheelID]->addChild(vehicleModel.wheelModel[wheelID].get());
		wheel[wheelID]->setActivationState(DISABLE_DEACTIVATION);	

		float	wheelFriction = 0.8f;//BT_LARGE_FLOAT;
		float	rollInfluence = 1.0f;//1.0f;

		btWheelInfo& wheelInfo = vehicle->getWheelInfo(wheelID);
		wheelInfo.m_suspensionStiffness = frontSuspensionStiffness;
		wheelInfo.m_wheelsDampingRelaxation = frontDampingRelaxationConstant * 2.0 * btSqrt(frontSuspensionStiffness);
		wheelInfo.m_wheelsDampingCompression = frontDampingCompressionConstant * 2.0 * btSqrt(frontSuspensionStiffness);
		wheelInfo.m_frictionSlip = wheelFriction;
		wheelInfo.m_rollInfluence = rollInfluence;
	}
	else{

		vehicle->addWheel(connectionPointCS0, wheelDirectionCS0, wheelAxleCS, rearSuspensionRestLength, rearWheelRadius, tuning, isFrontWheel);

		wheel[wheelID] = physics->createRigidBody(world, 20, tr, vehicleModel.wheelShape[wheelID]);
		wheelPAT[wheelID] = new osg::PositionAttitudeTransform;

		wheelPAT[wheelID]->addChild(vehicleModel.wheelModel[wheelID].get());
		wheel[wheelID]->setActivationState(DISABLE_DEACTIVATION);	

		float	wheelFriction = 0.8f;//BT_LARGE_FLOAT;
		float	rollInfluence = 1.0f;//1.0f;

		btWheelInfo& wheelInfo = vehicle->getWheelInfo(wheelID);
		wheelInfo.m_suspensionStiffness = rearSuspensionStiffness;
		wheelInfo.m_wheelsDampingRelaxation = rearDampingRelaxationConstant * 2.0 * btSqrt(rearSuspensionStiffness);
		wheelInfo.m_wheelsDampingCompression = rearDampingCompressionConstant * 2.0 * btSqrt(rearSuspensionStiffness);
		wheelInfo.m_frictionSlip = wheelFriction;
		wheelInfo.m_rollInfluence = rollInfluence;
	}
};

void Vehicle::bindKeys(int mode){

	if(keyEH->IsKeyPressed(osgGA::GUIEventAdapter::KEY_F1, true))
		driving = ACKERMAN;
	else if(keyEH->IsKeyPressed(osgGA::GUIEventAdapter::KEY_F2, true))
		driving = REVERSE_ACKERMAN;
	else if(keyEH->IsKeyPressed(osgGA::GUIEventAdapter::KEY_F3, true))
		driving = PARALLEL;
	else if(keyEH->IsKeyPressed(osgGA::GUIEventAdapter::KEY_F4, true))
		driving = HYBRID;

	if(mode & EXPERIMENTAL_ANALOG){
		if(keyEH->IsKeyDown('1')){
			steeringClamp = 0.8*0.1;
		}
		else if(keyEH->IsKeyDown('2')){
			steeringClamp = 0.8*0.2;
		}
		else if(keyEH->IsKeyDown('3')){
			steeringClamp = 0.8*0.3;
		}
		else if(keyEH->IsKeyDown('4')){
			steeringClamp = 0.8*0.4;
		}
		else if(keyEH->IsKeyDown('5')){
			steeringClamp = 0.8*0.1;
		}
		else if(keyEH->IsKeyDown('6')){
			steeringClamp = 0.8*0.6;
		}
		else if(keyEH->IsKeyDown('7')){
			steeringClamp = 0.8*0.7;
		}
		else if(keyEH->IsKeyDown('8')){
			steeringClamp = 0.8*0.8;
		}
		else if(keyEH->IsKeyDown('9')){
			steeringClamp = 0.8*0.9;
		}
		else
			steeringClamp = 0.8;
	}

	if(keyEH->IsKeyDown(keys.throttle)){
		if(mode & EXPERIMENTAL_ANALOG){
			if(keyEH->IsKeyDown('o')){
				throttle = 0.1;
			}
			else if(keyEH->IsKeyDown('i')){
				throttle = 0.2;
			}
			else if(keyEH->IsKeyDown('u')){
				throttle = 0.3;
			}
			else if(keyEH->IsKeyDown('y')){
				throttle = 0.4;
			}
			else if(keyEH->IsKeyDown('t')){
				throttle = 0.1;
			}
			else if(keyEH->IsKeyDown('r')){
				throttle = 0.6;
			}
			else if(keyEH->IsKeyDown('e')){
				throttle = 0.7;
			}
			else if(keyEH->IsKeyDown('w')){
				throttle = 0.8;
			}
			else if(keyEH->IsKeyDown('q')){
				throttle = 0.9;
			}
			else
				throttle = 1.0;
		}
		else
			throttle = 1;
	}
	else
		throttle = 0;

	if(keyEH->IsKeyDown(keys.brake)){
		brake = 1;
		findNodeByName(vehicleModel.chassisModel.get(),"brzda_on")->setNodeMask(0xffffffff);
		findNodeByName(vehicleModel.chassisModel.get(),"brzda_off")->setNodeMask(0x0);
	}
	else{
		brake = 0;
		findNodeByName(vehicleModel.chassisModel.get(),"brzda_on")->setNodeMask(0x0);
		findNodeByName(vehicleModel.chassisModel.get(),"brzda_off")->setNodeMask(0xffffffff);
	}

	if(keyEH->IsKeyDown(keys.right)){
		if(gVehicleSteering >= -0.2)
			gVehicleSteering -= steeringIncrement/2.0;
		else
			gVehicleSteering -= steeringIncrement;
		if (gVehicleSteering < -steeringClamp)
			gVehicleSteering = -steeringClamp;
	}
	else if(keyEH->IsKeyDown(keys.left)){
		if(gVehicleSteering <= 0.2)
			gVehicleSteering += steeringIncrement/2.0;
		else
			gVehicleSteering += steeringIncrement;
		if (gVehicleSteering > steeringClamp)
			gVehicleSteering = steeringClamp;
	}
	else{
		if(fabs(gVehicleSteering) <= steeringIncrement)
			gVehicleSteering = 0;
		else if(gVehicleSteering < 0)
			gVehicleSteering += steeringIncrement;
		else if(gVehicleSteering > 0)
			gVehicleSteering -= steeringIncrement;
	}
	
	if(keyEH->IsKeyPressed(keys.reset)){
		resetVehicle();
	}

	if(keyEH->IsKeyDown(keys.handbrake)){
		handbrake = true;
	}
	else{
		handbrake = false;	
	}

	if(keyEH->IsKeyPressed(keys.gear_up) && gear < gear_ratio.size()-1 && (mode & DISABLE_GEARBOX) == 0){
		gear++;
		if(gear > 0)
			findNodeByName(vehicleModel.chassisModel.get(),"rev_light")->setNodeMask(0x0);
	}
	
	if(keyEH->IsKeyPressed(keys.gear_down) && gear > 0 && (mode & DISABLE_GEARBOX) == 0){
		if((engine_rpm / gear_ratio.at(gear)) * gear_ratio.at(gear-1) < (highest_rpm + 2000))
			gear--;

		if(gear == 0)
			findNodeByName(vehicleModel.chassisModel.get(),"rev_light")->setNodeMask(0xffffffff);
	}

	if(keyEH->IsKeyDown(keys.clutch)){
		clutch = 1;
	}
	else{
		clutch = 0;	
	}
};

void Vehicle::updateVehicle(int seed, float dt, int mode){
	random_seed = seed;

	updateVehicle(dt, mode);
}

void Vehicle::updateVehicle(float dt, int mode){
	float wheel_rpm;

	if(vehicle->getCurrentSpeedKmHour() < 10 && throttle == 0)
		clutch = 1;
	else
		clutch = 0;

	if(mode != DISABLE_ALL_KEYS)
		bindKeys(mode);
	else
		handbrake = 1;


///////////////	

	if( manual == false ){
		if( gear == 1 && throttle > 0 )
			gear++;
		else if( gear == 2 && throttle == 0 && vehicle->getCurrentSpeedKmHour() < 0.5 )
			gear--;
		else if( gear == 1 && vehicle->getCurrentSpeedKmHour() < 0.5 && brake == 1 ){
			gear--;
			findNodeByName(vehicleModel.chassisModel.get(),"rev_light")->setNodeMask(0xffffffff);
		}
		else if( gear == 0 && vehicle->getCurrentSpeedKmHour() > -0.5 && throttle == 1 ){
			gear++;
			findNodeByName(vehicleModel.chassisModel.get(),"rev_light")->setNodeMask(0x0);
		}
		else if( gear == 0 && brake == 1 ){
			brake = 0;
			throttle = 1;
			findNodeByName(vehicleModel.chassisModel.get(),"brzda_on")->setNodeMask(0x0);
			findNodeByName(vehicleModel.chassisModel.get(),"brzda_off")->setNodeMask(0xffffffff);
		}
		else if( gear == 0 && throttle == 1 ){
			brake = 1;
			throttle = 0;
			findNodeByName(vehicleModel.chassisModel.get(),"brzda_on")->setNodeMask(0xffffffff);
			findNodeByName(vehicleModel.chassisModel.get(),"brzda_off")->setNodeMask(0x0);
		}
	}

	if(gear != 1 && clutch != 1){
		// wheel rpm		
		wheel_rpm = fabs(((vehicle->getCurrentSpeedKmHour()/0.06) / (3.141592654 * 2 * frontWheelRadius)));

		// engine rpm
		engine_rpm = fabs(wheel_rpm * diff_ratio * gear_ratio.at(gear));

		if(engine_rpm < lowest_rpm)
			engine_rpm = lowest_rpm;

		float torque;

		if(engine_rpm > highest_rpm){
			torque = -1 * GetTorque(highest_rpm);
			throttle = 1;
		}
		else
			torque = GetTorque(engine_rpm);

	////////////////
		if(gear > 1 && throttle == 0 && brake == 0 && vehicle->getCurrentSpeedKmHour() > 0.1)
			throttle = -0.5;
		else if(gear == 0  && throttle == 0 && brake == 0 && vehicle->getCurrentSpeedKmHour() < -0.1)
			throttle = -0.5;
		gEngineForce = throttle * torque * diff_ratio * gear_ratio.at(gear) / 0.35;// - air_tunnel * (0.5 * 0.39 * chassisHeight * chassisWidth * 1.164 * (vehicle->getCurrentSpeedKmHour()/3.6) * fabs(vehicle->getCurrentSpeedKmHour()/3.6));

		if(throttle == -0.5)
			throttle = 0;

		if(drive == 1){
			if(!handbrake){
				vehicle->applyEngineForce(gEngineForce/2.0, 3);
				vehicle->applyEngineForce(gEngineForce/2.0, 2);
			}
			else{
				vehicle->applyEngineForce(0, 3);
				vehicle->applyEngineForce(0, 2);
			}
		}
		else if(drive == 2){
			if(!handbrake){
				vehicle->applyEngineForce(gEngineForce/4.0, 3);
				vehicle->applyEngineForce(gEngineForce/4.0, 2);
			}
			else{
				vehicle->applyEngineForce(0, 3);
				vehicle->applyEngineForce(0, 2);
			}
			vehicle->applyEngineForce(gEngineForce/4.0, 1);
			vehicle->applyEngineForce(gEngineForce/4.0, 0);
		}
		else{
			vehicle->applyEngineForce(gEngineForce/2.0, 1);
			vehicle->applyEngineForce(gEngineForce/2.0, 0);
		}

		if( manual == false ){
			if( engine_rpm >= highest_rpm && gear > 1)
				gear++;
			else if( gear > 2 ){
				engine_rpm = wheel_rpm * diff_ratio * gear_ratio.at((gear-1));	
				if( engine_rpm <= highest_rpm - 500 )
					gear--;
			}
		}

	}
	else{
		vehicle->applyEngineForce(0, 3);
		vehicle->applyEngineForce(0, 2);
		vehicle->applyEngineForce(0, 1);
		vehicle->applyEngineForce(0, 0);
	}

	if(brake == 0)
		gBreakingForce = 0;
	else
		gBreakingForce = maxBrakingForce;

	vehicle->setBrake(gBreakingForce, 0);
	vehicle->setBrake(gBreakingForce, 1);

	float af = -(0.5 * dragCoefficient * chassisHeight * chassisWidth * physics->getAirDensity() * (vehicle->getCurrentSpeedKmHour()/3.6) * fabs(vehicle->getCurrentSpeedKmHour()/3.6));
	btVector3 airForce(af,af,af);
	chassis->applyCentralForce(airForce*vehicle->getForwardVector());

	// to the right
	if(gVehicleSteering < 0){

		if( driving == HYBRID ){
			float l = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getY() - vehicle->getWheelInfo(3).m_chassisConnectionPointCS.getY();
			float w = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getX()*2;
			
			// reverse ackerman mechanism
			if( vehicle->getCurrentSpeedKmHour() > 20.0 ){
				vehicle->setSteeringValue(gVehicleSteering, 1); // left
				gVehicleSteering *= -1;
				vehicle->setSteeringValue(-1*(acot((cos(gVehicleSteering)/sin(gVehicleSteering)) - w/l)), 0); // right
				gVehicleSteering *= -1;
			}
			// ackerman mechanism
			else{
				vehicle->setSteeringValue(gVehicleSteering, 0); // right
				gVehicleSteering *= -1;
				vehicle->setSteeringValue(-1*(acot(w/l + (cos(gVehicleSteering)/sin(gVehicleSteering)))), 1); // left
				gVehicleSteering *= -1;
			}
		}
		else if( driving == ACKERMAN ){
			float l = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getY() - vehicle->getWheelInfo(3).m_chassisConnectionPointCS.getY();
			float w = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getX()*2;
		
			vehicle->setSteeringValue(gVehicleSteering, 0); // right
			gVehicleSteering *= -1;
			vehicle->setSteeringValue(-1*(acot(w/l + (cos(gVehicleSteering)/sin(gVehicleSteering)))), 1); // left
			gVehicleSteering *= -1;
			
		}
		else if( driving == REVERSE_ACKERMAN ){
			float l = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getY() - vehicle->getWheelInfo(3).m_chassisConnectionPointCS.getY();
			float w = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getX()*2;
		
			vehicle->setSteeringValue(gVehicleSteering, 1); // left
			gVehicleSteering *= -1;
			vehicle->setSteeringValue(-1*(acot((cos(gVehicleSteering)/sin(gVehicleSteering)) - w/l)), 0); // right
			gVehicleSteering *= -1;
			
		} 
		else{ // parallel
			vehicle->setSteeringValue(gVehicleSteering, 0); // right
			vehicle->setSteeringValue(gVehicleSteering, 1); // left
		}
	}
	// to the left
	else if(gVehicleSteering > 0){

		if( driving == HYBRID ){
			float l = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getY() - vehicle->getWheelInfo(3).m_chassisConnectionPointCS.getY();
			float w = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getX()*2;
			
			// reverse ackerman mechanism
			if( vehicle->getCurrentSpeedKmHour() > 20.0 ){
				vehicle->setSteeringValue(gVehicleSteering, 0); // right
				vehicle->setSteeringValue(acot((cos(gVehicleSteering)/sin(gVehicleSteering)) - w/l), 1); // left
			}
			// ackerman mechanism
			else{
				vehicle->setSteeringValue(gVehicleSteering, 1); // left
				vehicle->setSteeringValue(acot(w/l + (cos(gVehicleSteering)/sin(gVehicleSteering))), 0); // right
			}
		}
		else if( driving == ACKERMAN ){
			float l = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getY() - vehicle->getWheelInfo(3).m_chassisConnectionPointCS.getY();
			float w = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getX()*2;
		
			vehicle->setSteeringValue(gVehicleSteering, 1); // left
			vehicle->setSteeringValue(acot(w/l + (cos(gVehicleSteering)/sin(gVehicleSteering))), 0); // right	
		}
		else if( driving == REVERSE_ACKERMAN ){
			float l = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getY() - vehicle->getWheelInfo(3).m_chassisConnectionPointCS.getY();
			float w = vehicle->getWheelInfo(0).m_chassisConnectionPointCS.getX()*2;
		
			vehicle->setSteeringValue(gVehicleSteering, 0); // right
			vehicle->setSteeringValue(acot((cos(gVehicleSteering)/sin(gVehicleSteering)) - w/l), 1); // left
		} 
		else{ // parallel
			vehicle->setSteeringValue(gVehicleSteering, 0); // right
			vehicle->setSteeringValue(gVehicleSteering, 1); // left
		}
	}
	// straight
	else{
		vehicle->setSteeringValue(gVehicleSteering, 0); // right
		vehicle->setSteeringValue(gVehicleSteering, 1); // left
	}

	if(handbrake){
		//cout<<maxHandbrakeForce<<endl;
		//maxHandbrakeForce = 10000;
		vehicle->setBrake(maxHandbrakeForce, 2);
		vehicle->setBrake(maxHandbrakeForce, 3);		
	}
	else{
		vehicle->setBrake(0, 2);
		vehicle->setBrake(0, 3);		
	}

	btScalar m[16];
	int i;

	btVector3	worldBoundsMin,worldBoundsMax;
	world->getBroadphase()->getBroadphaseAabb(worldBoundsMin,worldBoundsMax);

	for (i=0;i<vehicle->getNumWheels();i++)
	{
		//synchronize the wheels with the (interpolated) chassis worldtransform
		vehicle->updateWheelTransform(i,true);
		//draw wheels (cylinders)
		vehicle->getWheelInfo(i).m_worldTransform.getOpenGLMatrix(m);

		osg::Matrixf mat(m);

		wheelPAT[i]->setPosition(mat.getTrans());
		wheelPAT[i]->setAttitude(mat.getRotate());
	}	

	// edit engine rpm for tachometer
	if(clutch == 1 || gear == 1){
		if(engine_rpm <= lowest_rpm){
			srand(random_seed);
			engine_rpm = rand() % 100 + lowest_rpm - 50;
		}
		else if(engine_rpm >= highest_rpm && engine_rpm <= highest_rpm + 100){
			srand(random_seed);
			engine_rpm = rand() % 100 + highest_rpm - 50;
		}
		else{
			if(throttle == 0){
				if(engine_rpm - 100 > lowest_rpm)
					engine_rpm -= 100;
			}
			else{
				if(engine_rpm + 100 < highest_rpm)
					engine_rpm += 100;
			}
		}
	}
	else{
		engine_rpm = fabs(wheel_rpm * diff_ratio * gear_ratio.at(gear));
	}

	if(engine_rpm >= highest_rpm && engine_rpm <= highest_rpm + 100){
		srand(random_seed);
		engine_rpm = rand() % 100 + highest_rpm - 50;
	}

	// smooth arrow rotation
	if(tachometer_rpm < engine_rpm){
		tachometer_rpm += 100;
		if(tachometer_rpm > engine_rpm)
			tachometer_rpm = engine_rpm;
	}
	else if(tachometer_rpm > engine_rpm){
		tachometer_rpm -= 200;
		if(tachometer_rpm < engine_rpm)
			tachometer_rpm = engine_rpm;
	}

};

osg::ref_ptr<osg::PositionAttitudeTransform> Vehicle::getChassisPAT(){
	return chassisPAT;	
};

osg::ref_ptr<osg::PositionAttitudeTransform> Vehicle::getWheelPAT(int wheelID){
	return wheelPAT[wheelID];	
};

// parsing ini file
void Vehicle::parseConfig(const char* configFile){
	string str;						// line in string
	string::size_type comment;		// position of comment char #
	char line[256];					// line in char[]
	char pom_str[256],pom_str2[256];// pomocny string
	float pom_int;					// pomocny int
	ifstream confFile(configFile);	// konfiguracny subor
	int inDnsTable = 0;				// flag, ci nacitavame dns tabulku
	string gr;						// pomocny string pre interface

	float hp_peak = -1;
	float torque_peak = -1;
  
	
	hp_peak_rpm = -1;
	torque_peak_rpm = -1;
	lowest_rpm = -1;
	highest_rpm = -1;
	// nacitavanie celeho suboru po riadokch
	while( confFile.getline(line, 256) ){
    
		// ulozenie riadku do 'string'u kvoli niektorym operaciam
		str.assign(line);
    
		// odstranenie whitespaces zo zaciatku riadku
		//str = lTrim(str);
    
		// zistenie pozicie komentara
		comment = str.find_first_of('#');
    
		// ak sa komentar nasiel,tak sa vsetko za nim odstrani
		if(comment < 256)
		str.erase(comment);
    
		// ak mame neprazdny string, tak naplnime strukturu dnsConfig
		if(str.length()>0){
			// test na interfaces
			if( str.find( "gearRatios ", 0 ) != string::npos ){
	
				//odstranenie "gear_ratios "
				for(int i=strlen("gearRatios "); i < (int)str.size(); i++){	
					// ak narazime na whitespace, tak sa ulozi iface do struktury
					if(isspace(str.at(i))){
						// ak je vsetko v poriadku, tak sa iface ulozi
						if((int)gr.size() > 0)
							gear_ratio.push_back(atof(gr.c_str()));
						gr.clear();
					}
					// zaznamenanie znaku
					else
						gr.push_back(str.at(i));
				}
			
				// ak nebola na konci medzera, tak treba pridat ten,co je ulozeny
				if(gr.size()>0){
					// ak je vsetko v poriadku, tak sa iface ulozi
//					if((int)gr.size() > 0)
						gear_ratio.push_back(atof(gr.c_str()));
					gr.clear();
				}
				continue;
			}

			// test na interfaces
			if( str.find( "frontWheelChassisConnectionPoint ", 0 ) != string::npos ){
				pom_int = 0;
				//odstranenie "gear_ratios "
				for(int i=strlen("frontWheelChassisConnectionPoint "); i < (int)str.size(); i++){	
					// ak narazime na whitespace, tak sa ulozi iface do struktury
					if(isspace(str.at(i))){
						// ak je vsetko v poriadku, tak sa iface ulozi
						if((int)gr.size() > 0){
							if(pom_int == 0)
								frontWheelChassisConnectionPoint.setX(atof(gr.c_str()));
							if(pom_int == 1)
								frontWheelChassisConnectionPoint.setY(atof(gr.c_str()));
							if(pom_int == 2)
								frontWheelChassisConnectionPoint.setZ(atof(gr.c_str()));
						}
						gr.clear();
						pom_int++;
					}
					// zaznamenanie znaku
					else
						gr.push_back(str.at(i));
				}
			
				// ak nebola na konci medzera, tak treba pridat ten,co je ulozeny
				if(gr.size()>0){
					if(pom_int == 0)
						frontWheelChassisConnectionPoint.setX(atof(gr.c_str()));
					if(pom_int == 1)
						frontWheelChassisConnectionPoint.setY(atof(gr.c_str()));
					if(pom_int == 2)
						frontWheelChassisConnectionPoint.setZ(atof(gr.c_str()));
					gr.clear();
				}
				continue;
			}

			// test na interfaces
			if( str.find( "rearWheelChassisConnectionPoint ", 0 ) != string::npos ){
				pom_int = 0;
				//odstranenie "gear_ratios "
				for(int i=strlen("rearWheelChassisConnectionPoint "); i < (int)str.size(); i++){	
					// ak narazime na whitespace, tak sa ulozi iface do struktury
					if(isspace(str.at(i))){
						// ak je vsetko v poriadku, tak sa iface ulozi
						if((int)gr.size() > 0){
							if(pom_int == 0)
								rearWheelChassisConnectionPoint.setX(atof(gr.c_str()));
							if(pom_int == 1)
								rearWheelChassisConnectionPoint.setY(atof(gr.c_str()));
							if(pom_int == 2)
								rearWheelChassisConnectionPoint.setZ(atof(gr.c_str()));
						}
						gr.clear();
						pom_int++;
					}
					// zaznamenanie znaku
					else
						gr.push_back(str.at(i));
				}
			
				// ak nebola na konci medzera, tak treba pridat ten,co je ulozeny
				if(gr.size()>0){
					if(pom_int == 0)
						rearWheelChassisConnectionPoint.setX(atof(gr.c_str()));
					if(pom_int == 1)
						rearWheelChassisConnectionPoint.setY(atof(gr.c_str()));
					if(pom_int == 2)
						rearWheelChassisConnectionPoint.setZ(atof(gr.c_str()));
					gr.clear();
				}
				continue;
			}
      
			// test na dns tabulku
			if( str.find( "dynoGraph", 0 ) != string::npos ){
				inDnsTable = 1;
				continue;
			}
      
			// sme v dns tabulke
			if(inDnsTable == 1){
				// test ci sa nejedna o {, alebo }
				pom_int = -1;
				float rpm,hp,trq; // kontrola tvaru IP
				if(sscanf(str.c_str(), "%f %f %f",&rpm, &hp, &trq) < 3){
					// jedna sa o {,}, alebo nejaky bordel...tak otestujeme ci je }, ak ano,tak sme vonku z dns tabulky
					if( str.find( "}", 0 ) != string::npos )
						inDnsTable = 0;
				}
				// mame aspon dva povinne udaje, tak doplnime do struktury
				else{
					hp_curve[rpm] = hp;
					torque_curve[rpm] = trq;
					if(hp_peak_rpm == -1 || (hp_peak <= hp)){
						hp_peak_rpm = rpm;
						hp_peak = hp;
					}
					if(torque_peak_rpm == -1 || (torque_peak <= trq)){
						torque_peak_rpm = rpm;
						torque_peak = trq;
					}
					if(lowest_rpm == -1 || lowest_rpm > rpm)
						lowest_rpm = rpm;
					if(highest_rpm == -1 || highest_rpm < rpm)
						highest_rpm = rpm;
				}
			}
			// sme mimo dns tabulky
			else if(inDnsTable == 0){
				if(sscanf(str.c_str(), "frontSuspensionRestLength %f", &pom_int) == 1)
					frontSuspensionRestLength = pom_int;
				else if(sscanf(str.c_str(), "rearSuspensionRestLength %f", &pom_int) == 1)
					rearSuspensionRestLength = pom_int;
				else if(sscanf(str.c_str(), "maxFrontSuspensionTravel %f", &pom_int) == 1)
					maxFrontSuspensionTravel = pom_int;
				else if(sscanf(str.c_str(), "maxRearSuspensionTravel %f", &pom_int) == 1)
					maxRearSuspensionTravel = pom_int;
				else if(sscanf(str.c_str(), "frontWheelRadius %f", &pom_int) == 1)
					frontWheelRadius = pom_int;
				else if(sscanf(str.c_str(), "rearWheelRadius %f", &pom_int) == 1)
					rearWheelRadius = pom_int;
				else if(sscanf(str.c_str(), "frontWheelWidth %f", &pom_int) == 1)
					frontWheelWidth = pom_int;
				else if(sscanf(str.c_str(), "rearWheelWidth %f", &pom_int) == 1)
					rearWheelWidth = pom_int;
				else if(sscanf(str.c_str(), "frontSuspensionStiffness %f", &pom_int) == 1)
					frontSuspensionStiffness = pom_int;
				else if(sscanf(str.c_str(), "rearSuspensionStiffness %f", &pom_int) == 1)
					rearSuspensionStiffness = pom_int;
				else if(sscanf(str.c_str(), "frontDampingCompressionConstant %f", &pom_int) == 1)
					frontDampingCompressionConstant = pom_int;
				else if(sscanf(str.c_str(), "rearDampingCompressionConstant %f", &pom_int) == 1)
					rearDampingCompressionConstant = pom_int;
				else if(sscanf(str.c_str(), "frontDampingRelaxationConstant %f", &pom_int) == 1)
					frontDampingRelaxationConstant = pom_int;
				else if(sscanf(str.c_str(), "rearDampingRelaxationConstant %f", &pom_int) == 1)
					rearDampingRelaxationConstant = pom_int;
				else if(sscanf(str.c_str(), "mass %f", &pom_int) == 1)
					mass = pom_int;
				else if(sscanf(str.c_str(), "length %f", &pom_int) == 1)
					chassisLength = pom_int;
				else if(sscanf(str.c_str(), "width %f", &pom_int) == 1)
					chassisWidth = pom_int;
				else if(sscanf(str.c_str(), "height %f", &pom_int) == 1)
					chassisHeight = pom_int;
				else if(sscanf(str.c_str(), "drive %f", &pom_int) == 1)
					drive = pom_int;
				else if(sscanf(str.c_str(), "reverseRatio %f", &pom_int) == 1)
					gear_ratio.at(0) = (-1)*pom_int;
				else if(sscanf(str.c_str(), "differentialRatio %f", &pom_int) == 1)
					diff_ratio = pom_int;
				else if(sscanf(str.c_str(), "dragCoefficient %f", &pom_int) == 1)
					dragCoefficient = pom_int;
				else if(sscanf(str.c_str(), "maxBrakingForce %f", &pom_int) == 1)
					maxBrakingForce = pom_int;
				else if(sscanf(str.c_str(), "maxHandbrakeForce %f", &pom_int) == 1)
					maxHandbrakeForce = pom_int;
			}
		}
  }
//  cout<<lowest_rpm<<" "<<highest_rpm<<" "<<hp_peak_rpm<<" "<<torque_peak_rpm<<endl;
};

float Vehicle::GetTorque(float rpm){
	
	// return 0 if rpm is not defined
	if(rpm < lowest_rpm || rpm > highest_rpm)
		return 0;

	map<float,float>::iterator up,down;
	float rate,trq_int;

	// get equal of higher rpm
	up = torque_curve.lower_bound(rpm);

	// get lower rpm
	down = up;
	if((*down).first == lowest_rpm)
		return (*up).second;
	
	down._Dec();

	// calculate rate between down and up
	rate = ((rpm - (*down).first)/((*up).first - (*down).first)) * 3.1415927;
	trq_int = (1 - cos(rate)) * 0.5;

	return (*down).second*(1-trq_int) + (*up).second*trq_int;
};

float Vehicle::GetHorsepower(float rpm){
	
	// return 0 if rpm is not defined
	if(rpm < lowest_rpm || rpm > highest_rpm)
		return 0;

	map<float,float>::iterator up,down;
	float rate,hp_int;

	// get equal of higher rpm
	up = hp_curve.lower_bound(rpm);

	// get lower rpm
	down = up;
	if((*down).first == lowest_rpm)
		return (*up).second;
	
	down._Dec();

	// calculate rate between down and up
	rate = ((rpm - (*down).first)/((*up).first - (*down).first)) * 3.1415927;
	hp_int = (1 - cos(rate)) * 0.5;

	return (*down).second*(1-hp_int) + (*up).second*hp_int;
};

char Vehicle::getGear(){
	if(gear == 0)
		return 'R';
	else if(gear == 1)
		return 'N';

	return '0'+gear-1;	
}

float Vehicle::getSpeed(){
	return vehicle->getCurrentSpeedKmHour();
}

int Vehicle::getRPM(){
	return engine_rpm;
}

int Vehicle::getTachometerRPM(){
	return tachometer_rpm;
}

VehicleHUDData Vehicle::getHudData(){
	VehicleHUDData data;
	data.gear = getGear();
	data.RPM = tachometer_rpm/1000.0f;
	data.speed = vehicle->getCurrentSpeedKmHour();

	osg::Vec3d pos = chassisPAT.get()->getPosition();
	osg::Quat rot = chassisPAT.get()->getAttitude();
	data.pos = osg::Vec4d(pos.x(),pos.y(),rot.z(),rot.w());

	data.driving = driving;

	return data;
}

void Vehicle::getHudDebugData(){

}

void Vehicle::setFriction(const int wheelID, float friction){
	btWheelInfo& wheelInfo = vehicle->getWheelInfo(wheelID);
	wheelInfo.m_frictionSlip = friction;

	//if(wheelID == 0)
	//	cout<<friction<<" ";
	//else if(wheelID == 1)
	//	cout<<friction<<endl;
}

void Vehicle::setAirTunnel(float airTunnel){
	air_tunnel = airTunnel;
}

float Vehicle::getAirTunnel(){
	return air_tunnel;
}