/*
	Physical Simulation in Graphics Scene - master's thesis
    Copyright (C) 2010-2011  Marin Javorka, FIT VUT in Brno, http://www.fit.vutbr.cz, marian.javorka@deq.sk

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/
#ifndef _VEHICLE_H
#define _VEHICLE_H

#include "btBulletDynamicsCommon.h"
#include "BulletDynamics/Vehicle/btRaycastVehicle.h"
#include "LinearMath/btQuickprof.h"
#include "LinearMath/btGeometryUtil.h"
#include "BulletCollision/CollisionShapes/btShapeHull.h"

#include <osg/PositionAttitudeTransform>
#include <osgDB/ReadFile>

#include <vector>
#include <iostream>
#include <map>

#include "physics.h"
#include "my_key_event_handler.h"

#include "time.h"


struct KeyMap{
	int throttle;
	int brake;
	int left;
	int right;
	int reset;
	int handbrake;
	int clutch;
	int gear_up;
	int gear_down;
};

struct VehicleModel{
	osg::ref_ptr<osg::Node> chassisModel;						// chassis model
	osg::ref_ptr<osg::Node> wheelModel[4];						// wheel model
	btCollisionShape* chassisShape;							// chassis collision shape
	btCollisionShape* wheelShape[4];							// wheel collision shape
};

struct VehicleHUDData{
	float RPM;
	char gear;
	float speed;
	osg::Vec4d pos;
	int driving;
};

struct VehicleHUDDebugData{
	
};

class Vehicle{
	public:
		enum UPDATE_MODES{UPDATE_ALL = 0x0, DISABLE_GEARBOX = 0x01, DISABLE_ALL_KEYS = 0x02, EXPERIMENTAL_ANALOG = 0x04};
		enum DRIVING_SYSTEM{HYBRID = 0, ACKERMAN, REVERSE_ACKERMAN, PARALLEL};

	private:
		btRaycastVehicle::btVehicleTuning tuning;					// vehicle tuning
		btVehicleRaycaster* rayCaster;								// ray caster
		//btAlignedObjectArray<btCollisionShape*> collsisionShapes;	// array of collision shapes
		
		KeyMap keys;
		VehicleModel vehicleModel;

		btRigidBody* chassis;										// chassis rigid body
		osg::ref_ptr<osg::Node> chassisModel;						// chassis model
		osg::ref_ptr<osg::PositionAttitudeTransform> chassisPAT;	// chassis PAT
		//btConvexHullShape* chassisShape;							// chassis collision shape
		btCollisionShape* chassisShape;							// chassis collision shape

		btRigidBody* wheel[4];										// wheel rigid body
		osg::ref_ptr<osg::Node> wheelModel[4];						// wheel model
		osg::ref_ptr<osg::PositionAttitudeTransform> wheelPAT[4];	// wheel PAT
		btCollisionShape* wheelShape[4];							// wheel collision shape
		btVector3 wheelDirectionCS0;								// direction of ray cast (to the ground)
		btVector3 wheelAxleCS;										// axle of rotation (X)

		// vehicle tuning info
		float mass;													// vehicle mass
		float chassisLength, chassisWidth, chassisHeight;			// chassis dimensions
		short int drive;											// drive type 0 - FWD, 1 - RWD, 2 - 4WD
		vector<float> gear_ratio;									// gear ratios
		float diff_ratio;											// differetial ratio
		map<float,float> hp_curve;									// engine power curve
		map<float,float> torque_curve;								// engine torque curve
		float hp_peak_rpm;											// power peak
		float torque_peak_rpm;										// torque peak
		float lowest_rpm;											// lowest possible rpm for the engine
		float highest_rpm;											// highest possible rpm for the engine (rev-limitter)
		btVector3 frontWheelChassisConnectionPoint;					// center of right front wheel
		btVector3 rearWheelChassisConnectionPoint;					// center of right rear wheel
		float frontSuspensionRestLength;							// maximum length of the front suspension
		float rearSuspensionRestLength;								// maximum length of the rear suspension
		float maxFrontSuspensionTravel;								// maximum front suspension compression
		float maxRearSuspensionTravel;								// maximum rear suspension compression
		float frontSuspensionStiffness;								// stiffness constant for the front suspension
		float rearSuspensionStiffness;								// stiffness constant for the rear suspension
		float frontDampingCompressionConstant;						// damping compression constant for the front suspension
		float rearDampingCompressionConstant;						// damping compression constant for the rear suspension
		float frontDampingRelaxationConstant;						// damping relaxation constant for the front suspension
		float rearDampingRelaxationConstant;						// damping relaxation constant for the rear suspension
		float dragCoefficient;										// coefficient of drag
		float maxBrakingForce;										// maximal braking force
		float maxHandbrakeForce;									// max handbrake force

		// vehicle variables		
		float gBreakingForce;
		float gVehicleSteering;				// current vehicle steering
		float steeringIncrement;			// steer increment
		float steeringClamp;				// maximal steer
		float throttle;						// pressure on throttle <0,1>
		float brake;						// pressure on brake
		bool handbrake;						// flag if handbrake is used
		int gear;							// actual gear
		int engine_rpm;						// actual engine RPM
		int tachometer_rpm;					// actual RPM on tachometer
		int clutch;
		//float	maxEngineForce;//this should be engine/velocity dependent

		int random_seed;

		float air_tunnel;

		int driving;



	osg::ref_ptr<MyKeyEventHandler> handler;


	btTransform tr;
	btDynamicsWorld* world;

	osg::Vec3f resetPosition;
		// private functions
		void createWheel(const btVector3 &connectionPointCS0, bool isFrontWheel, int wheelID);

		// bind keys
		void bindKeys(int mode);

		void updateVehicle(float dt, int mode);

		void parseConfig(const char* configFile);

		float GetHorsepower(float rpm); 
		
		float GetTorque(float rpm); 

	public:
		float gEngineForce;					// amount of applied torque on wheel	
		bool manual;
		float frontWheelRadius;										// front wheel radius
		float rearWheelRadius;										// rear wheel radius
		float frontWheelWidth;										// front wheel width
		float rearWheelWidth;										// rear wheel width

		btRaycastVehicle* vehicle;									// vehicle
		// constructor
		Vehicle(btDynamicsWorld* m_world, osg::Vec3f pos, VehicleModel model, KeyMap keyMap, bool get_manual = false);
		// destructor
		~Vehicle();

		// vehicle configuration
		void configVehicle(const char* configFile);

		// reset vehicle
		void resetVehicle();

		// create vehicle
		void createVehicle();

		void updateVehicle(int seed, float dt, int mode);

		osg::ref_ptr<osg::PositionAttitudeTransform> getChassisPAT();
		osg::ref_ptr<osg::PositionAttitudeTransform> getWheelPAT(int wheelID);

		btRigidBody* getChassis() {return chassis;}

		btRigidBody* getWheel(int id) {return wheel[id];}

		char getGear();

		float getSpeed();

		int getRPM();

		int getTachometerRPM();

		VehicleHUDData getHudData();
		
		void getHudDebugData();

		void setFriction(const int wheelID, float friction = 0.8);

		void setAirTunnel(float airTunnel);
		float getAirTunnel();

		static osg::Node* findNodeByName(osg::Node* node,const std::string& name);

		static float acot(float val){ return 3.12/2.0 - atan(val); }
};

#endif